from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit


def simple_urljoin(*parts, append_slash=False):
    """Normalize url parts and join them with a slash."""
    parts = list(map(str, parts))
    schemes, netlocs, paths, queries, fragments = zip(*(urlsplit(part) for part in parts))
    scheme = _last(schemes)
    netloc = _last(netlocs)
    paths = [x for x in paths if x]
    if scheme and not netloc and paths:
        netloc, *paths = paths
    path = '/'.join((x.strip('/') for x in paths if x.strip('/')))

    if paths and parts[0].startswith('/'):
        path = '/' + path
    if append_slash or (paths and parts[-1].endswith('/')):
        path += '/'
    query = _last(queries)
    fragment = _last(fragments)
    return urlunsplit((scheme, netloc, path, query, fragment))


def _last(sequence, default=''):
    not_empty = [x for x in sequence if x]
    return not_empty[-1] if not_empty else default


def add_query(url: str, **params: str):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    query_params.update(params)
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))



from operator import attrgetter
from typing import Hashable, Iterable, List, Sized


class classproperty:
    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)


def is_empty(val):
    """
        Check where value is logically `empty` - does not contain information.
        False and 0 are not considered empty, but empty collections are.
    """
    if val is None or isinstance(val, Sized) and len(val) == 0:  # Empty string is also Sized of len 0
        return True
    return False


def unique_ordered(sequence: Iterable[Hashable]) -> List:
    return list(dict.fromkeys(sequence))


def unique_objs(objs: List[object], unique_attrs: List[str]) -> List:
    """
       Get list of unique objs from sequence,
        preserving order when the objs first occurred in original sequence.
    """

    seen_obj_footprints = set()
    unique = []
    footprint_func = attrgetter(*unique_attrs)
    for obj in objs:
        obj_footprint = footprint_func(obj)
        if obj_footprint in seen_obj_footprints:
            continue

        seen_obj_footprints.add(obj_footprint)
        unique.append(obj)
    return unique



from django.core.exceptions import EmptyResultSet
from django.db import models
from django.db.models import Avg
from django.db.models import BooleanField
from django.db.models import Count
from django.db.models import ExpressionWrapper
from django.db.models import Max
from django.db.models import Min
from django.db.models import Subquery
from django.db.models import Sum
from django.db.models import Value


class BooleanQ(ExpressionWrapper):
    output_field = BooleanField()

    def __init__(self, *args, **kwargs):
        expression = models.Q(*args, **kwargs)
        super().__init__(expression, output_field=None)

    def as_sql(self, compiler, connection):
        try:
            return super().as_sql(compiler, connection)
        except EmptyResultSet:
            return compiler.compile(Value(False))


class SubqueryAgg(Subquery):
    template = '(SELECT %(func)s(%(field_name)s) FROM (%(subquery)s) _sub)'

    def __init__(self, queryset, field_name, func, **kwargs):
        queryset = queryset.order_by()
        if not queryset.query.values_select:
            queryset = queryset.values(field_name)
        super().__init__(queryset, field_name=field_name, func=func, **kwargs)


class SubqueryAvg(SubqueryAgg):
    def __init__(self, queryset, field_name, **kwargs):
        super().__init__(queryset, field_name, func=Avg.function, **kwargs)


class SubqueryMin(SubqueryAgg):
    def __init__(self, queryset, field_name, **kwargs):
        super().__init__(queryset, field_name, func=Min.function, **kwargs)


class SubqueryMax(SubqueryAgg):
    def __init__(self, queryset, field_name, **kwargs):
        super().__init__(queryset, field_name, func=Max.function, **kwargs)


class SubquerySum(SubqueryAgg):
    def __init__(self, queryset, field_name, **kwargs):
        super().__init__(queryset, field_name, func=Sum.function, **kwargs)


class SubqueryCount(SubqueryAgg):
    output_field = models.IntegerField()

    def __init__(self, queryset, **kwargs):
        if not queryset.query.values_select:
            queryset = queryset.values('pk')
        super().__init__(queryset, field_name='*', func=Count.function, **kwargs)



import contextlib
import logging
from functools import wraps
from typing import Callable, Iterable, Mapping

from django_redis import get_redis_connection
from redis.exceptions import LockError


logger = logging.getLogger(__name__)

NO_CACHE = object()
DEFAULT_LOCK_TIMEOUT = 10
KeyMakerType = Callable[[Iterable, Mapping], str]


def redis_client(write=True):
    return get_redis_connection(write=write)


@contextlib.contextmanager
def use_lock(key, timeout=DEFAULT_LOCK_TIMEOUT, blocking_timeout=None):
    """
    If blocking_timeout is specified, will proceed after
    blocking_timeout even without the lock obtained.
    Check the return value and abort if needed.
    """

    if timeout is None:  # Prevent locks without timeout set - it is too dangerous if the process dies
        timeout = DEFAULT_LOCK_TIMEOUT

    lock = redis_client().lock(f'use_lock:{key}', timeout=timeout, blocking_timeout=blocking_timeout)
    acquired = lock.acquire(blocking=True)
    try:
        yield acquired
    finally:
        if not acquired:
            return

        # Don't want failed lock release to break the whole application
        try:
            lock.release()
        except LockError as exc:
            logger.exception(str(exc))


def with_lock(
    key,
    timeout,
    blocking_timeout=None,
    proceed_without_lock=False,
):
    """
    If blocking_timeout is specified, and the lock is not obtained after blocking_timeout,
    behavior will depend on the proceed_without_lock.
    """

    def factory(func):
        @wraps(func)
        def decorator(*args, **kwargs):
            with use_lock(key, timeout, blocking_timeout) as acquired:
                if not (acquired or proceed_without_lock):
                    return None
                return func(*args, **kwargs)

        return decorator

    return factory


def ensure_single(key, timeout):
    """Ensure function executes only single time simultaneously, other executions are aborted immediately"""
    return with_lock(key, timeout, blocking_timeout=0, proceed_without_lock=False)



from typing import Iterable
from typing import List
from typing import Type

from django.db import models


def get_update_fields(cls: Type[models.Model], exclude: Iterable[str] = None) -> List[str]:
    exclude = set(exclude) if exclude is not None else {}

    def _can_be_updated(field):
        return (
            field.name not in exclude and
            not (field.many_to_many or field.one_to_many)
            and not field.auto_created
        )

    return [
        field.name for field in cls._meta.get_fields()
        if _can_be_updated(field)
    ]
